# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An example of parsing full TPTP CNF subset."""
import json
import logging
import os
import sys
from glob import glob
from typing import Optional

from tptp_lark_parser import TPTPParser


def _read_and_parse_file(
    tptp_parser: TPTPParser, problem_filename: str
) -> None:
    with open(problem_filename, "r", encoding="utf-8") as problem_file:
        problem_text = problem_file.read()
    tptp_parser.parse(problem_text)


def _get_logger(level: int) -> logging.Logger:
    logger = logging.getLogger()
    logger.setLevel(level)
    handler = logging.StreamHandler(sys.stdout)
    logger.addHandler(handler)
    return logger


def parse_tptp(
    tptp_folder: str,
    output_file: str,
    logging_level: int,
    tokens_filename: Optional[str],
    learn_new_tokens: bool,
) -> None:
    """
    Parse all TPTP CNF problems and write all symbols encountered.

    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> tptp_folder = (
    ...     files("tptp_lark_parser")
    ...     .joinpath(os.path.join("resources", "TPTP-mock"))
    ... )
    >>> temp_folder = getfixture("tmp_path")  # noqa: F821
    >>> output_file = os.path.join(temp_folder, "test.json")
    >>> os.path.exists(output_file)
    False
    >>> parse_tptp(tptp_folder, output_file, logging.FATAL, None, True)
    >>> parse_tptp(tptp_folder, output_file, logging.FATAL, output_file, False)

    :param tptp_folder: a folder with TPTP dump
    :param output_file: where to save tokens found
    :param logging_level: what to log
    :param tokens_filename: a filename of known tokens storage
    :param learn_new_tokens: if ``False``, then parsing fails when encounters
        an unknown token
    """
    logger = _get_logger(logging_level)
    tptp_parser = TPTPParser(tptp_folder, learn_new_tokens, tokens_filename)
    cnf_problems = sorted(
        [
            problem
            for problem in glob(
                os.path.join(tptp_folder, "Problems", "*", "*-*.p")
            )
            if "CSR" not in problem
            and "HWV" not in problem
            and "KRS" not in problem
            and "PLA" not in problem
            and "SWV" not in problem
            and "SYN" not in problem
            and "SYO" not in problem
        ]
    )
    for problem in cnf_problems:
        _read_and_parse_file(tptp_parser, problem)
        # we save lists on every iteration because the next one may fail
        tokens = {
            "variables": list(tptp_parser.cnf_parser.variable_map.keys()),
            "functions": list(tptp_parser.cnf_parser.function_map.keys()),
            "predicates": list(tptp_parser.cnf_parser.predicate_map.keys()),
        }
        with open(output_file, "w", encoding="utf-8") as tokens_file:
            json.dump(tokens, tokens_file)
        logger.info("%s done", problem)


if __name__ == "__main__":
    parse_tptp(
        os.path.join(os.environ["WORK"], "data", "TPTP-v8.0.0"),
        "tptp_tokens.json",
        logging.INFO,
        None,
        True,
    )  # pragma: no cover



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa D205
"""
Grammar.
********
"""
from dataclasses import dataclass, field
from typing import Optional, Tuple, Union
from uuid import uuid1

FALSEHOOD_SYMBOL = "$false"
FALSEHOOD_SYMBOL_ID = 0
EQUALITY_SYMBOL = "="
EQUALITY_SYMBOL_ID = 1
INEQUALITY_SYMBOL = "!="
INEQUALITY_SYMBOL_ID = 2


@dataclass(frozen=True)
class Variable:
    """
    A variable is characterised only by its name.

    .. _variable:
    """

    name: int


@dataclass(frozen=True)
class Function:
    """
    A functional symbol might be applied to a list of arguments.

    .. _Function:
    """

    name: int
    arguments: Tuple[Union[Variable, "Function"], ...]


Term = Union[Variable, Function]
Term.__doc__ = """
.. _Term:

Term is either a :ref:`Variable <Variable>` or a :ref:`Function <Function>`
"""


@dataclass(frozen=True)
class Predicate:
    """
    A predicate symbol might be applied to a list of arguments.

    .. _Predicate:
    """

    name: int
    arguments: Tuple[Term, ...]


Proposition = Union[Predicate, Term]
Proposition.__doc__ = """
.. _Proposition:

Proposition is either a :ref:`Predicate <Predicate>` or a :ref:`Term <Term>`
"""


@dataclass(frozen=True)
class Literal:
    """
    Literal is an atom which can be negated or not.

    .. _Literal:
    """

    negated: bool
    atom: Predicate


def _term_to_tptp(term: Term) -> str:
    if isinstance(term, Function):
        arguments = tuple(
            _term_to_tptp(argument) for argument in term.arguments
        )
        if arguments != tuple():
            return f"f{term.name}({','.join(arguments)})"
        return f"f{term.name}"
    return f"X{term.name}"


def _literal_to_tptp(literal: Literal) -> str:
    res = "~" if literal.negated else ""
    arguments = tuple(_term_to_tptp(term) for term in literal.atom.arguments)
    if literal.atom.name != EQUALITY_SYMBOL_ID:
        if literal.atom.name != FALSEHOOD_SYMBOL_ID:
            res += f"p{literal.atom.name}({','.join(arguments)})"
        else:
            res += f"{FALSEHOOD_SYMBOL}()"
    else:
        res += f"{arguments[0]} {EQUALITY_SYMBOL} {arguments[1]}"
    return res


@dataclass(frozen=True)
class Clause:
    """
    Clause is a disjunction of literals.

    .. _Clause:


    :param literals: a list of literals, forming the clause
    :param label: comes from the problem file or starts with ``inferred_`` if
         inferred during the episode
    :param role: formula role (axiom, hypothesis, etc)
    :param inference_parents: a list of labels from which the clause was
         inferred. For clauses from the problem statement, this list is empty
    :param inference_rule: the rule according to which the clause was got from
         the ``inference_parents``
    :param processed: boolean value splitting clauses into unprocessed and
         processed ones; in the beginning, everything is not processed
    :param birth_step: a number of the step when the clause appeared in the
         unprocessed set; clauses from the problem have ``birth_step`` zero
    """

    literals: Tuple[Literal, ...]
    label: str = field(
        default_factory=lambda: "x" + str(uuid1()).replace("-", "_")
    )
    role: str = "lemma"
    inference_parents: Optional[Tuple[str, ...]] = None
    inference_rule: Optional[str] = None
    processed: Optional[bool] = None
    birth_step: Optional[int] = None

    def __repr__(self):
        """Print a logical forumla back to TPTP language."""
        res = f"cnf({self.label}, {self.role}, "
        for literal in self.literals:
            res += _literal_to_tptp(literal) + " | "
        if res[-2:] == "| ":
            res = res[:-3]
        if not self.literals:
            res += FALSEHOOD_SYMBOL
        if (
            self.inference_parents is not None
            and self.inference_rule is not None
        ):
            res += (
                f", inference({self.inference_rule}, [], ["
                + ", ".join(self.inference_parents)
                + "])"
            )
        return res + ")."



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa D205
"""
CNF Parser.
===========
"""
import dataclasses
import json
from typing import Dict, Optional, Tuple, Union

from lark import Transformer

from tptp_lark_parser.grammar import (
    EQUALITY_SYMBOL,
    EQUALITY_SYMBOL_ID,
    FALSEHOOD_SYMBOL,
    FALSEHOOD_SYMBOL_ID,
    INEQUALITY_SYMBOL,
    INEQUALITY_SYMBOL_ID,
    Clause,
    Function,
    Literal,
    Predicate,
    Variable,
)


def _load_token_lists(
    tokens_filename: str,
) -> Tuple[Dict[str, int], Dict[str, int], Dict[str, int]]:
    with open(tokens_filename, "r", encoding="utf-8") as tokens_file:
        tokens = json.load(tokens_file)
    return (
        {v: k for k, v in enumerate(tokens["variables"])},
        {v: k for k, v in enumerate(tokens["functions"])},
        {v: k for k, v in enumerate(tokens["predicates"])},
    )


class CNFParser(Transformer):
    """
    A parser for ``<cnf_formula>`` from Lark parser tree.

    Methods are not typed since nobody calls them directly.

    >>> import sys
    >>> import os
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> from lark import Lark
    >>> lark_parser = Lark(
    ...     files("tptp_lark_parser")
    ...     .joinpath(os.path.join("resources", "TPTP.lark"))
    ...     .read_text(),
    ...     start="tptp_file"
    ... )
    >>> lark_parsed_tree = lark_parser.parse('''
    ...    cnf(this_is_a_test_case, axiom, f4(X1,f1(X2),f3(X3,f2)) = f4(X1,X2,f5)
    ...    | ~ p2(f4(X1),f1(X2)) | $false | p3,
    ...    inference(resolution, [], [this, that, third])).
    ... ''')
    >>> cnf_parser = CNFParser()
    >>> cnf_parser.function_map
    {'$not#a&function^': 0}
    >>> cnf_parser.predicate_map
    {'$false': 0, '=': 1, '!=': 2}
    >>> cnf_parser.variable_map
    {'X': 0}
    >>> cnf_parser.transform(lark_parsed_tree)
    Traceback (most recent call last):
    ...
    lark.exceptions.VisitError: Error trying to process rule "variable":
    <BLANKLINE>
    unknown symbol: X1
    >>> cnf_parser.extendable = True
    >>> cnf_parser.transform(lark_parsed_tree)
    cnf(this_is_a_test_case, axiom, f4(X1,f1(X2),f3(X3,f2)) = f4(X1,X2,f5) | ~p3(f4(X1),f1(X2)) | $false() | p4(), inference(resolution, [], [this, that, third])).
    >>> cnf_parser.function_map
    {'$not#a&function^': 0, 'f1': 1, 'f2': 2, 'f3': 3, 'f4': 4, 'f5': 5}
    >>> cnf_parser.predicate_map
    {'$false': 0, '=': 1, '!=': 2, 'p2': 3, 'p3': 4}
    """

    def __init__(
        self, tokens_filename: Optional[str] = None, extendable: bool = False
    ):
        """
        Initialize functional and predicate symbols lists.

        :param tokens_filename: a filename of known tokens storage
        :param extendable: when set to ``False``, the parser fails
            when encounters new symbols
        """
        super().__init__()
        if tokens_filename is None:
            self.function_map: Dict[str, int] = {"$not#a&function^": 0}
            self.predicate_map: Dict[str, int] = {
                FALSEHOOD_SYMBOL: FALSEHOOD_SYMBOL_ID,
                EQUALITY_SYMBOL: EQUALITY_SYMBOL_ID,
                INEQUALITY_SYMBOL: INEQUALITY_SYMBOL_ID,
            }
            self.variable_map: Dict[str, int] = {"X": 0}
        else:
            (
                self.variable_map,
                self.function_map,
                self.predicate_map,
            ) = _load_token_lists(tokens_filename)
        self.extendable = extendable

    def __default_token__(self, token):
        """All the tokens we return as is."""
        return token.value

    def __default__(self, data, children, meta):
        """Pop one element lists.

        By default, if we see a list of only one element, we return that
        element, not the whole list.
        """
        if len(children) == 1:
            return children[0]
        return children

    def _function(self, children):
        """Functional symbol with arguments."""
        function_name = children[0]
        function_id = self._get_symbol_id(function_name, Function)
        if len(children) > 1:
            return Function(function_id, tuple(children[1]))
        return Function(function_id, ())

    def _get_symbol_id(
        self, symbol: str, symbol_type: Union[Variable, Function, Predicate]
    ) -> int:
        """
        Get an integer ID for a symbol.

        :param symbol: a predicate or functional symbol, or a variable name
        :returns: an integer ID
        """
        if symbol_type == Variable:
            symbol_map = self.variable_map
        else:
            symbol_map = (
                self.function_map
                if symbol_type == Function
                else self.predicate_map
            )
        if symbol not in symbol_map:
            if self.extendable:
                symbol_map.update({symbol: 1 + max(symbol_map.values())})
            else:
                raise ValueError(f"unknown symbol: {symbol}")
        return symbol_map[symbol]

    def fof_defined_plain_formula(self, children):
        """
        Predicate with or without arguments.

        <fof_defined_plain_formula> :== <defined_proposition> | <defined_predicate>(<fof_arguments>)
        """
        return self._predicate(children)

    def fof_plain_term(self, children):
        """
        Functional symbol with or without (a constant) arguments.

        <fof_plain_term>       ::= <constant> | <functor>(<fof_arguments>)
        """
        return self._function(children)

    def fof_defined_term(self, children):
        """
        Another way to describe a functional symbol.

        <fof_defined_term>     ::= <defined_term> | <fof_defined_atomic_term>
        """
        return self._function(children)

    def variable(self, children):
        """
        Variable (supposed to be universally quantified).

        <variable>             ::= <upper_word>
        """
        return Variable(self._get_symbol_id(children[0], Variable))

    @staticmethod
    def fof_arguments(children):
        """
        List of arguments, organised in pairs.

        <fof_arguments>        ::= <fof_term> | <fof_term>,<fof_arguments>
        """
        result = ()
        for item in children:
            if not isinstance(item, (Variable, Function)):
                result = result + tuple(item)
            else:
                result = result + (item,)
        return result

    @staticmethod
    def literal(children):
        """
        Literal is a possible negated predicate.

        <literal>              ::= <fof_atomic_formula> | ~ <fof_atomic_formula> | <fof_infix_unary>
        """
        if children[0] == "~":
            return Literal(True, children[1])
        if isinstance(children[0], Predicate):
            if children[0].name == INEQUALITY_SYMBOL_ID:
                return Literal(
                    negated=True,
                    atom=Predicate(
                        name=EQUALITY_SYMBOL_ID,
                        arguments=children[0].arguments,
                    ),
                )
        return Literal(False, children[0])

    def _predicate(self, children):
        """Predicates are atomic formulae."""
        predicate_id = self._get_symbol_id(children[0], Predicate)
        if len(children) > 1:
            return Predicate(predicate_id, tuple(children[1]))
        return Predicate(predicate_id, ())

    def fof_plain_atomic_formula(self, children):
        """
        Another way for writing predicates.

        <fof_plain_atomic_formula> :== <proposition> | <predicate>(<fof_arguments>)
        """
        return self._predicate(children)

    def fof_defined_infix_formula(self, children):
        """
        Translte predicates in the infix form to the prefix.

        <fof_defined_infix_formula> ::= <fof_term> <defined_infix_pred> <fof_term>
        """
        predicate_id = self._get_symbol_id(children[1], Predicate)
        return Predicate(predicate_id, (children[0], children[2]))

    def fof_infix_unary(self, children):
        """
        Translate predicates in the infix form to the prefix.

        <fof_infix_unary>      ::= <fof_term> <infix_inequality> <fof_term>
        """
        predicate_id = self._get_symbol_id(children[1], Predicate)
        return Predicate(predicate_id, (children[0], children[2]))

    @staticmethod
    def disjunction(children):
        """
        Clause structure.

        <disjunction>          ::= <literal> | <disjunction> <vline> <literal>
        """
        if len(children) == 1:
            if (
                children[0].atom.name == FALSEHOOD_SYMBOL_ID
                and not children[0].negated
            ):
                return Clause(tuple())
            return Clause(tuple(children))
        literals = ()
        for item in (children[0], children[2]):
            if isinstance(item, Clause):
                literals = literals + item.literals
            else:
                literals = literals + (item,)
        return Clause(literals)

    @staticmethod
    def cnf_annotated(children):
        """
        Annotated CNF formula (clause).

        <cnf_annotated>        ::= cnf(<name>,<formula_role>,<cnf_formula> <annotations>).
        """
        clause = children[2]
        inference_rule = None
        inference_parents = None
        if isinstance(children[3], list):
            for annotation in children[3]:
                if isinstance(annotation, dict):
                    if "inference_record" in annotation:
                        inference_rule = annotation["inference_record"][0]
                        inference_parents = annotation["inference_record"][1]
        return dataclasses.replace(
            clause,
            label=children[0],
            role=children[1],
            inference_rule=inference_rule,
            inference_parents=inference_parents,
        )

    @staticmethod
    def inference_record(children):
        """
        Inference record (parents and rule).

        <inference_record>     :== inference(<inference_rule>,<useful_info>,
        <inference_parents>)
        """
        return {"inference_record": (children[0], tuple(children[2]))}

    @staticmethod
    def annotations(children):
        """
        Annotation (we care only about inference info from it).

        <annotations>          ::= ,<source><optional_info> | <null>
        """
        if isinstance(children, list):
            if len(children) == 1:
                return tuple(children[0])
        return children

    @staticmethod
    def parent_info(children):
        """
        Inference parents.

        <parent_info>          :== <source><parent_details>
        """
        return children[0]

    @staticmethod
    def parent_list(children):
        """
        Inference parents lits.

        <parent_list>          :== <parent_info> | <parent_info>,<parent_list>
        """
        if len(children) == 2:
            return (children[0],) + tuple(children[1])
        return children



# Copyright 2022 Boris Shminke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa D205
"""
TPTP Parser.
============
"""
import os
import sys
from typing import Optional, Tuple

from lark import Lark, Token

from tptp_lark_parser.cnf_parser import CNFParser
from tptp_lark_parser.grammar import Clause

if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    # pylint: disable=no-name-in-module, import-error
    from importlib.resources import files  # type: ignore
else:  # pragma: no cover
    from importlib_resources import files  # pylint: disable=import-error


# pylint: disable=too-few-public-methods
class TPTPParser:
    r"""
    TPTP parser.

    .. _tptp-parser:

    >>> tptp_folder = (
    ...     files("tptp_lark_parser")
    ...     .joinpath(os.path.join("resources", "TPTP-mock"))
    ... )
    >>> from tptp_lark_parser.grammar import (Literal, Predicate, Variable,
    ...     Function, EQUALITY_SYMBOL_ID)
    >>> tptp_parser = TPTPParser(
    ...     tptp_folder=tptp_folder, extendable=True, tokens_filename=None
    ... )
    >>> clause = Clause(label="this_is_a_test_case", literals=(Literal(True, Predicate(EQUALITY_SYMBOL_ID, (Function(1, (Variable(1), )), Variable(2)))), Literal(False, Predicate(EQUALITY_SYMBOL_ID, (Function(2, ()), Function(3, ())))), Literal(False, Predicate(3, (Variable(1),)))), inference_rule="resolution", inference_parents=("one", "two"))
    >>> tptp_parser.parse(str(clause))[0] == clause
    True
    >>> empty_clause = Clause(literals=())
    >>> tptp_parser.parse(str(empty_clause))[0] == empty_clause
    True
    >>> tptp_text = (
    ...     files("tptp_lark_parser")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST", "TST001-1.p"
    ...     ))
    ...     .read_text()
    ... )
    >>> parsed_clauses = tptp_parser.parse(tptp_text)
    >>> print("\n".join(map(str, parsed_clauses)))
    cnf(this_is_a_test_case_1, hypothesis, p4(f4), inference(resolution, [], [one, two])).
    cnf(this_is_a_test_case_2, hypothesis, ~p4(f4)).
    cnf(test_axiom, axiom, f4 = f5).
    cnf(test_axiom_2, axiom, ~f4 = f6).
    """

    _tokens_from_resources = str(
        files("tptp_lark_parser").joinpath(
            os.path.join("resources", "tptp_tokens.json")
        )
    )

    def __init__(
        self,
        tptp_folder: str = ".",
        extendable: bool = False,
        tokens_filename: Optional[str] = _tokens_from_resources,
    ):
        """
        Create a parser.

        We create a Lark parser based on the grammar file from package's
        resources.

        :param tptp_folder: a folder containing TPTP database
        :param extendable: when set to ``False``, the parser fails
            when encounters new symbols
        :param tokens_filename: a filename of known tokens storage
        """
        # pylint: disable=unspecified-encoding
        self.parser = Lark(
            files("tptp_lark_parser")
            .joinpath(os.path.join("resources", "TPTP.lark"))
            .read_text(),
            start="tptp_file",
        )
        self.tptp_folder = tptp_folder
        self.cnf_parser = CNFParser(tokens_filename, extendable)

    def parse(self, tptp_text: str) -> Tuple[Clause, ...]:
        """
        Recursively parse a string containing a TPTP problem.

        :param tptp_text: a name of a problem (or axioms) file
        :returns: a list of clauses (including those of the axioms)
        """
        problem_tree = self.parser.parse(tptp_text)
        clauses = tuple(
            self.cnf_parser.transform(cnf_formula)
            for cnf_formula in problem_tree.find_data("cnf_annotated")
        )
        for include in problem_tree.find_data("include"):
            token = include.children[0]
            if isinstance(token, Token):
                with open(
                    os.path.join(
                        self.tptp_folder, token.value.replace("'", "")
                    ),
                    "r",
                    encoding="utf-8",
                ) as included_file:
                    clauses = clauses + self.parse(included_file.read())
        return clauses
